/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions:	Routines to generate Knowledge Management rule and
**		procedure for cascade delete.
**
*/

/* Headers */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef IGNORE
#define IGNORE 9
#endif

#ifndef DB_OBJ_LEN
#define DB_OBJ_LEN 24
#endif

#ifndef DB_OBJ_LEN_S
#define DB_OBJ_LEN_S 25
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif



/* Function Prototypes */

int cascade_del_dbproc();
int casc_del_dbproc(char *);
int casc_del_rule(char *, char *);
void casc_del_dbp_hdr();
void casc_del_dbproc_args();
void casc_del_sql();
void casc_del_auto(char *);
void casc_del_man(char *, int);
void final_casc_del();
void cascade_del_rule();
void casc_del_rule_hdr(char *);
void casc_del_rule_action();
void casc_del_rule_args();

extern int check_existing(char *, char *, char *, char *);
extern void get_name(char *, char *, char *, int, char *);
extern int run_ddl(char *);
extern void set_type_length(int, char *, char *);
extern void list_tables();


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

extern struct obj_name
{
char name[33];
};


/* Global Variables */

FILE *DDL_ptr;

extern int No_key_cols, No_auto_relns, No_man_relns;

extern char Table_name[33], Dbproc_name[DB_OBJ_LEN_S];
extern char Rdate[25], Username[25];

extern struct obj_name *Relns_auto, *Relns_man, *Man_cols;
extern struct key_info *Keyptr;



/* FUNCTION CODE */


/*
** Generate a cascade delete database procedure, if necessary.
** The main logic of this DELETE function is that if a row in a 'Parent'
** Table is deleted, then rows with corresponding Foreign keys
** in related tables must also be deleted.
*/

int cascade_del_dbproc()
{
    int result;
    char ddl_script[FILE_LEN];

    get_name(Table_name, "\0", "_casc_del", DB_OBJ_LEN, Dbproc_name);
    if (check_existing(Table_name, Dbproc_name, ddl_script, "dbproc") == IGNORE)
	return TRUE;

    printf("\n\nCreating database procedure for Cascade Delete ...");
    result = casc_del_dbproc(ddl_script);
    final_casc_del(result);

    return result;
}


/* Finalise */

void final_casc_del(result)
int result;
{
    if (result == TRUE)
	printf("\nCreated ok\n");
    else
	printf("\nDDL script created, but it contains errors\n");

    system("sleep 2");

    return;
}


/* Control creation of ddl sql script file for a DB Procedure */

int casc_del_dbproc(ddl_script)
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    casc_del_dbp_hdr();
    casc_del_dbproc_args();
    casc_del_sql();

    fprintf(DDL_ptr, "END;\n\\p\\g\n");
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void casc_del_dbp_hdr()
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Procedure:  %s\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.dbproc\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database Procedure to maintain referential "\
		     "integrity for tables that\n");
    fprintf(DDL_ptr, "**   have foreign keys to the %s table.\n", Table_name);
    fprintf(DDL_ptr, "**   This will cascade deletions in %s\n", Table_name);
    fprintf(DDL_ptr, "**   to foreign keys in the following tables:-\n");

    list_tables();

    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop procedure before creating  */\n\n");
    fprintf(DDL_ptr, "DROP PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create procedure %s  */\n\n", Dbproc_name);
    fprintf(DDL_ptr, "CREATE PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "(\n");

    return;
}


/* Write in the procedure arguments */

void casc_del_dbproc_args()
{
    int i;
    char c;
    char type_length[15], null_default[25];
    struct key_info *kptr;

    c = ',';

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;

	set_type_length(kptr->c_length, kptr->c_datatype, type_length);

	if (kptr->c_null[0] == 'Y')
	    strcpy(null_default, "with null");
	else
	    {
	    strcpy(null_default, "not null");
	    if (kptr->c_default[0] == 'Y')
		strcat(null_default, " with default");
	    else
		strcat(null_default, " not default");
	    }

	/* New name */

	if ((No_key_cols - i) == 1)
	    c = '\0';

	fprintf(DDL_ptr, "    %-32s  %-15s  %s%c\n", kptr->c_name, type_length,
						     null_default, c);
	}

    return;
}


/* Create the main logic */

void casc_del_sql()
{
    int i;
    struct obj_name *tmp_ptr;

    fprintf(DDL_ptr, ") =\n");
    fprintf(DDL_ptr, "BEGIN\n");

    for(i = 0; i < No_auto_relns; i++)
	{
	tmp_ptr = Relns_auto + i;
	casc_del_auto(tmp_ptr->name);
	}

    for(i = 0; i < No_man_relns; i++)
	{
	tmp_ptr = Relns_man + i;
	casc_del_man(tmp_ptr->name, i * No_key_cols);
	}

    return;
}


/* Add in the DELETE sql statement (automatic relations) */

void casc_del_auto(rel_name)
char *rel_name;
{
    int i;
    char c;
    struct key_info *kptr;

    fprintf(DDL_ptr, "\t/* Cascade deletion to %s table */\n", rel_name);
    fprintf(DDL_ptr, "\n\tDELETE FROM %s\n", rel_name);

    c = '\0';

    for(i = 0; i < No_key_cols; i++)
	{
	if ((No_key_cols - i) == 1)
	    c = ';';

	kptr = Keyptr + i;

	if (i == 0)
	    fprintf(DDL_ptr, "\tWHERE %s = :%s%c\n", kptr->c_name,
						     kptr->c_name, c);
	else
	    fprintf(DDL_ptr, "\tAND %s = :%s%c\n", kptr->c_name,
						   kptr->c_name, c);
	}

    fprintf(DDL_ptr, "\n");

    return;
}


/* Add in the DELETE sql statement (manual relations) */

void casc_del_man(rel_name, ref_pt)
char *rel_name;
int ref_pt;
{
    int i;
    char c;
    struct key_info *kptr;
    struct obj_name *mc_ptr;

    fprintf(DDL_ptr, "\t/* Cascade deletion to %s table */\n", rel_name);
    fprintf(DDL_ptr, "\n\tDELETE FROM %s\n", rel_name);

    c = '\0';

    for(i = 0; i < No_key_cols; i++)
	{
	mc_ptr = Man_cols + (ref_pt + i);

	if ((No_key_cols - i) == 1)
	    c = ';';

	kptr = Keyptr + i;

	if (i == 0)
	    fprintf(DDL_ptr, "\tWHERE %s = :%s%c\n", mc_ptr->name,
						     kptr->c_name, c);
	else
	    fprintf(DDL_ptr, "\tAND %s = :%s%c\n", mc_ptr->name,
						   kptr->c_name, c);
	}

    fprintf(DDL_ptr, "\n");

    return;
}


/*
** Generate a cascade delete database rule, if necessary.
** See cascade delete database procedure comments for details.
*/

void cascade_del_rule()
{
    int result;
    char rule_name[DB_OBJ_LEN_S];
    char ddl_script[FILE_LEN];

    get_name(Table_name, "\0", "_casc_del", DB_OBJ_LEN, rule_name);
    if (check_existing(Table_name, rule_name, ddl_script, "rule") == IGNORE)
	return;

    printf("\n\nCreating database rule for Cascade Delete ...");
    result = casc_del_rule(rule_name, ddl_script);
    final_casc_del(result);

    return;
}


/* Control creation of ddl sql script file for a DB Rule */

int casc_del_rule(rule_name, ddl_script)
char *rule_name;
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    casc_del_rule_hdr(rule_name);
    casc_del_rule_action();
    casc_del_rule_args();

    fprintf(DDL_ptr, "\\p\\g\n");
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void casc_del_rule_hdr(rule_name)
char *rule_name;
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Rule:  %s\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.rule\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database rule to enable referential "\
		     "integrity for tables that\n");
    fprintf(DDL_ptr, "**   have foreign keys to the %s table.\n", Table_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop rule before creating  */\n\n");
    fprintf(DDL_ptr, "DROP RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create Rule %s  */\n\n", rule_name);
    fprintf(DDL_ptr, "CREATE RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\n");

    return;
}


/* Set up the rule action clause */

void casc_del_rule_action()
{
    fprintf(DDL_ptr, "\tAFTER DELETE ON %s\n\n", Table_name);
    fprintf(DDL_ptr, "\tEXECUTE PROCEDURE %s(\n", Dbproc_name);

    return;
}


/* Format the arguments to the database procedure */

void casc_del_rule_args()
{
    int i;
    struct key_info *kptr;

    for(i = 0; i < (No_key_cols - 1); i++)
	{
	kptr = Keyptr + i;
	fprintf(DDL_ptr, "\t\t\t%s = old.%s,\n", kptr->c_name, kptr->c_name);
	}

    kptr = Keyptr + i;
    fprintf(DDL_ptr, "\t\t\t%s = old.%s)\n", kptr->c_name, kptr->c_name);

    return;
}
