/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions:	Routines to generate Knowledge Management rule and
**		procedure for cascade update.
**
*/

/* Headers */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef IGNORE
#define IGNORE 9
#endif

#ifndef DB_OBJ_LEN
#define DB_OBJ_LEN 24
#endif

#ifndef DB_OBJ_LEN_S
#define DB_OBJ_LEN_S 25
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif



/* Function Prototypes */

int cascade_upd_dbproc();
int casc_upd_dbproc(char *);
int casc_upd_rule(char *, char *);
void casc_upd_dbp_hdr();
void casc_upd_dbproc_args();
void casc_upd_sql();
void casc_upd_auto(char *);
void casc_upd_man(char *, int);
void final_casc_upd();
void cascade_upd_rule();
void casc_upd_rule_hdr(char *);
void casc_upd_rule_action();
void casc_upd_rule_args();

extern int check_existing(char *, char *, char *, char *);
extern void get_name(char *, char *, char *, int, char *);
extern int run_ddl(char *);
extern void main_final();
extern void set_type_length(int, char *, char *);
extern void list_tables();
extern void def_val_code(char *);


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

extern struct obj_name
{
char name[33];
};


/* Global Variables */

FILE *DDL_ptr;

struct obj_name *Old_names;

extern int No_key_cols, No_auto_relns, No_man_relns;
extern int Def_key_vals;

extern char Table_name[33], Dbproc_name[DB_OBJ_LEN_S];
extern char Rdate[25], Username[25];

extern struct obj_name *Relns_auto, *Relns_man, *Man_cols;
extern struct key_info *Keyptr;



/* FUNCTION CODE */


/*
** Generate a cascade update database procedure, if necessary.
** The main logic of this UPDATE function is that if the
** Primary key in a Table is changed, then the corresponding 
** Foreign keys in related tables must also be changed.
*/

int cascade_upd_dbproc()
{
    int result;
    char ddl_script[FILE_LEN];

    get_name(Table_name, "\0", "_casc_upd", DB_OBJ_LEN, Dbproc_name);
    if (check_existing(Table_name, Dbproc_name, ddl_script, "dbproc") == IGNORE)
	return TRUE;

    printf("\n\nCreating database procedure for Cascade Update ...");
    result = casc_upd_dbproc(ddl_script);
    final_casc_upd(result);

    return result;
}


/* Finalise */

void final_casc_upd(result)
int result;
{
    if (result == TRUE)
	printf("\nCreated ok\n");
    else
	printf("\nDDL script created, but it contains errors\n");

    system("sleep 2");

    return;
}


/* Control creation of ddl sql script file for a DB Procedure */

int casc_upd_dbproc(ddl_script)
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    casc_upd_dbp_hdr();
    casc_upd_dbproc_args();
    casc_upd_sql();

    fprintf(DDL_ptr, "END;\n\\p\\g\n");
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void casc_upd_dbp_hdr()
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Procedure:  %s\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.dbproc\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database Procedure to maintain referential "\
		     "integrity for tables that\n");
    fprintf(DDL_ptr, "**   have foreign keys to the %s table.\n", Table_name);
    fprintf(DDL_ptr, "**   This will cascade primary key changes in %s\n",
								  Table_name);
    fprintf(DDL_ptr, "**   to foreign keys in the following tables:-\n");

    list_tables();

    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop procedure before creating  */\n\n");
    fprintf(DDL_ptr, "DROP PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create procedure %s  */\n\n", Dbproc_name);
    fprintf(DDL_ptr, "CREATE PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "(\n");

    return;
}


/* Write in the procedure arguments */

void casc_upd_dbproc_args()
{
    int i;
    char c;
    char type_length[15], null_default[25];
    char tmp_name[33];
    struct key_info *kptr;
    struct obj_name *tmp_ptr;

    /* Make space for 'Old name' argument name */

    i = No_key_cols * 33;
    Old_names = (struct obj_name *) calloc(i, sizeof(struct obj_name));
    if (Old_names == NULL)
	{
	printf("\nERROR: No space available for old name arguments\n");
	main_final();
	exit(2);
	}

    c = ',';

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;

	set_type_length(kptr->c_length, kptr->c_datatype, type_length);

	if (kptr->c_null[0] == 'Y')
	    strcpy(null_default, "with null");
	else
	    {
	    strcpy(null_default, "not null");
	    if (kptr->c_default[0] == 'Y')
		strcat(null_default, " with default");
	    else
		strcat(null_default, " not default");
	    }

	/* New name */

	fprintf(DDL_ptr, "    %-32s  %-15s  %s%c\n", kptr->c_name, type_length,
						     null_default, c);

	/* Work out old name and save */

	get_name(kptr->c_name, "old_", "\0", DB_OBJ_LEN, tmp_name);
	tmp_ptr = Old_names + i;
	strcpy(tmp_ptr->name, tmp_name);

	if ((No_key_cols - i) == 1)
	    c = '\0';

	fprintf(DDL_ptr, "    %-32s  %-15s  %s%c\n", tmp_name, type_length,
						     null_default, c);
	}

    return;
}


/* Create the main logic */

void casc_upd_sql()
{
    int i;
    struct obj_name *tmp_ptr;

    fprintf(DDL_ptr, ") =\n");

    if (Def_key_vals != 3)
	{
	fprintf(DDL_ptr, "DECLARE\n");
	fprintf(DDL_ptr, "\terror_number =\tinteger4\tnot null with "\
			 "default;\n");
	fprintf(DDL_ptr, "\trows =\t\tinteger4\tnot null with default;\n");
	fprintf(DDL_ptr, "\tmsg_text =\tvarchar(240)\tnot null with "\
			 "default;\n\n");
	}

    fprintf(DDL_ptr, "BEGIN\n");

    if (Def_key_vals != 3)
        def_val_code(Table_name);

    for(i = 0; i < No_auto_relns; i++)
	{
	tmp_ptr = Relns_auto + i;
	casc_upd_auto(tmp_ptr->name);
	}

    for(i = 0; i < No_man_relns; i++)
	{
	tmp_ptr = Relns_man + i;
	casc_upd_man(tmp_ptr->name, i * No_key_cols);
	}

    return;
}


/* Add in the UPDATE sql statement (automatic relations) */

void casc_upd_auto(rel_name)
char *rel_name;
{
    int i;
    char c;
    struct key_info *kptr;
    struct obj_name *tmp_ptr;

    fprintf(DDL_ptr, "\n\t/* Cascade key change to %s table */\n", rel_name);
    fprintf(DDL_ptr, "\n\tUPDATE %s\n", rel_name);

    c = ',';

    for(i = 0; i < No_key_cols; i++)
	{
	if ((No_key_cols - i) == 1)
	    c = '\0';

	kptr = Keyptr + i;

	if (i == 0)
	    fprintf(DDL_ptr, "\tSET %s = :%s%c\n", kptr->c_name,
					           kptr->c_name, c);
	else
	    fprintf(DDL_ptr, "\t    %s = :%s%c\n", kptr->c_name,
					           kptr->c_name, c);
	}

    c = '\0';

    for(i = 0; i < No_key_cols; i++)
	{
	tmp_ptr = Old_names + i;

	if ((No_key_cols - i) == 1)
	    c = ';';

	kptr = Keyptr + i;

	if (i == 0)
	    fprintf(DDL_ptr, "\tWHERE %s = :%s%c\n", kptr->c_name,
						     tmp_ptr->name, c);
	else
	    fprintf(DDL_ptr, "\tAND %s = :%s%c\n", kptr->c_name,
						   tmp_ptr->name, c);
	}

    return;
}


/* Add in the UPDATE sql statement (manual relations) */

void casc_upd_man(rel_name, ref_pt)
char *rel_name;
int ref_pt;
{
    int i;
    char c;
    struct key_info *kptr;
    struct obj_name *on_ptr;
    struct obj_name *mc_ptr;

    fprintf(DDL_ptr, "\n\t/* Cascade key change to %s table */\n", rel_name);
    fprintf(DDL_ptr, "\n\tUPDATE %s\n", rel_name);

    c = ',';

    for(i = 0; i < No_key_cols; i++)
	{
	mc_ptr = Man_cols + (ref_pt + i);

	if ((No_key_cols - i) == 1)
	    c = '\0';

	kptr = Keyptr + i;

	if (i == 0)
	    fprintf(DDL_ptr, "\tSET %s = :%s%c\n", mc_ptr->name,
						   kptr->c_name, c);
	else
	    fprintf(DDL_ptr, "\t    %s = :%s%c\n", mc_ptr->name,
						   kptr->c_name, c);
	}

    c = '\0';

    for(i = 0; i < No_key_cols; i++)
	{
	mc_ptr = Man_cols + (ref_pt + i);
	on_ptr = Old_names + i;

	if ((No_key_cols - i) == 1)
	    c = ';';

	if (i == 0)
	    fprintf(DDL_ptr, "\tWHERE %s = :%s%c\n", mc_ptr->name,
						     on_ptr->name, c);
	else
	    fprintf(DDL_ptr, "\tAND %s = :%s%c\n", mc_ptr->name,
						   on_ptr->name, c);
	}

    return;
}


/*
** Generate a cascade update database rule, if necessary.
** See cascade update database procedure comments for details.
*/

void cascade_upd_rule()
{
    int result;
    char rule_name[DB_OBJ_LEN_S];
    char ddl_script[FILE_LEN];

    get_name(Table_name, "\0", "_casc_upd", DB_OBJ_LEN, rule_name);
    if (check_existing(Table_name, rule_name, ddl_script, "rule") == IGNORE)
	return;

    printf("\n\nCreating database rule for Cascade Update ...");
    result = casc_upd_rule(rule_name, ddl_script);
    final_casc_upd(result);

    return;
}


/* Control creation of ddl sql script file for a DB Rule */

int casc_upd_rule(rule_name, ddl_script)
char *rule_name;
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    casc_upd_rule_hdr(rule_name);
    casc_upd_rule_action();
    casc_upd_rule_args();

    fprintf(DDL_ptr, "\\p\\g\n");
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void casc_upd_rule_hdr(rule_name)
char *rule_name;
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Rule:  %s\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.rule\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database rule to enable referential "\
		     "integrity for tables that\n");
    fprintf(DDL_ptr, "**   have foreign keys to the %s table.\n", Table_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop rule before creating  */\n\n");
    fprintf(DDL_ptr, "DROP RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create Rule %s  */\n\n", rule_name);
    fprintf(DDL_ptr, "CREATE RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\n");

    return;
}


/* Set up the rule action clause */

void casc_upd_rule_action()
{
    int i;
    struct key_info *kptr;

    kptr = Keyptr;

    if (No_key_cols == 1)
	fprintf(DDL_ptr, "\tAFTER UPDATE (%s) ON %s\n\n", kptr->c_name,
							  Table_name);
    else
	{
	fprintf(DDL_ptr, "\tAFTER UPDATE ON %s\n", Table_name);
	fprintf(DDL_ptr, "\t\tWHERE new.%s != old.%s\n", kptr->c_name,
							 kptr->c_name);
	for(i = 1; i < No_key_cols; i++)
	    {
	    kptr = Keyptr + i;
	    fprintf(DDL_ptr, "\t\tOR new.%s != old.%s\n", kptr->c_name,
							  kptr->c_name);
	    }
	fprintf(DDL_ptr, "\n");
	}

    fprintf(DDL_ptr, "\tEXECUTE PROCEDURE %s(\n", Dbproc_name);

    return;
}


/* Format the arguments to the database procedure */

void casc_upd_rule_args()
{
    int i;
    char old_name[33];
    struct key_info *kptr;

    for(i = 0; i < (No_key_cols - 1); i++)
	{
	kptr = Keyptr + i;
	fprintf(DDL_ptr, "\t\t\t%s = new.%s,\n", kptr->c_name, kptr->c_name);
	get_name(kptr->c_name, "old_", "\0", DB_OBJ_LEN, old_name);
	fprintf(DDL_ptr, "\t\t\t%s = old.%s,\n", old_name, kptr->c_name);
	}

    kptr = Keyptr + i;
    fprintf(DDL_ptr, "\t\t\t%s = new.%s,\n", kptr->c_name, kptr->c_name);
    get_name(kptr->c_name, "old_", "\0", DB_OBJ_LEN, old_name);
    fprintf(DDL_ptr, "\t\t\t%s = old.%s)\n", old_name, kptr->c_name);

    return;
}
