/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions:	General routines
**
*/

/* Headers */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef IGNORE
#define IGNORE 9
#endif

#ifndef DB_OBJ_LEN
#define DB_OBJ_LEN 24
#endif

#ifndef DB_OBJ_LEN_S
#define DB_OBJ_LEN_S 25
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif

#ifndef BUFFER_LEN
#define BUFFER_LEN 100
#endif

#ifndef MAND_COL
#define MAND_COL "mandatory_col"
#endif

#define TMP_LEN 255


/* Function Prototypes */

int check_existing(char *, char *, char *, char *);
int km_object(char *, char *, char *, char *);
void get_name(char *, char *, char *, int, char *);
void trunc_string(char *, int);
void ddl_open(char *, char *);
void append_file(char *, char *);
void list_tables();
void def_val_code(char *);
void display_help();

extern int read_line(char *, int);
extern void main_final();
extern void set_def_value(char *, char *);


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

extern struct obj_name
{
char name[33];
};


/* Global Variables */

extern FILE *DDL_ptr;

extern int No_key_cols, No_auto_relns, No_man_relns;
extern int Def_key_vals;

extern char Dbname[33];
extern char Table_name[33];
extern char Default_path[FILE_LEN];
extern char Restriction[500];
extern char Username[25];

extern struct obj_name *Relns_auto, *Relns_man;
extern struct key_info *Keyptr;



/* FUNCTION CODE */


/* Check if a KM object already exists and determine the script file name */

int check_existing(tbl_name, obj_name, ddl_script, ext)
char *tbl_name;
char *obj_name;
char *ddl_script;
char *ext;
{
    int rc;

    /* Determine the name and see if it already exists */

    sprintf(ddl_script, "%s/%s/%s.%s", Default_path, tbl_name, obj_name, ext);

    rc = km_object(tbl_name, obj_name, ddl_script, ext);

    if (rc == IGNORE)
        return IGNORE;

    return TRUE;
}


/*
** Check if a Knowledge Management object (probably a Rule or Database
** Procedure exists and check if the creation script script file also
** exists. Detremine the appropriate action for combinations of what
** exists and what doesn't.
*/

int km_object(tbl_name, km_name, ddl_script, obj_type)
char *tbl_name;
char *km_name;
char *ddl_script;
char *obj_type;
{
    int i, j, rc, rc2;
    char opt;
    char tmp_str[TMP_LEN], obj_cat[DB_OBJ_LEN_S], obj_ext[10], obj_desc[20];
    struct stat stbuf;

    /* Type of object */

    if (strcmp(obj_type, "dbproc") == 0)
	{
	sprintf(Restriction,
		"procedure_name = '%s'\nand procedure_owner = '%s'",
		km_name, Username);
	strcpy(obj_desc, "database procedure");
	strcpy(obj_cat, "iiprocedures");
	strcpy(obj_ext, "dbproc");
	}

    else if (strcmp(obj_type, "rule") == 0)
	{
	sprintf(Restriction,
		"rule_name = '%s'\nand rule_owner = '%s'",
		km_name, Username);
	strcpy(obj_desc, "database rule");
	strcpy(obj_cat, "iirules");
	strcpy(obj_ext, "rule");
	}

    /* See if object already exists */

    rc = rc2 = FALSE;

    for(i = 2; rc != TRUE && rc2 != -1; i++)
	{
	rc = sql_lookup(Dbname, obj_cat);
	rc2 = stat(ddl_script, &stbuf);

	/* Both DB object and script file do not exist */

	if (rc == FALSE && rc2 == -1)
	    continue;

	/* DB object does not exist, script file exists */

	if ((rc == FALSE) &&
	    (rc2 == 0 && ((stbuf.st_mode & S_IFMT) == S_IFREG)))
	    {
	    sprintf(tmp_str, "%s.bak", ddl_script);
	    printf("\n\nWARNING: %s script file already exists.", obj_desc);
	    printf("\n         Renaming to %s ...", tmp_str);
	    rc2 = -1;

	    if (rename(ddl_script, tmp_str) != 0)
		{
		printf("\nERROR: Unable to rename file. Run aborts.\n");
		main_final();
		exit(1);
		}

	    continue;
	    }

	/* Both DB object and script file exist */

	if ((rc == TRUE) &&
	    (rc2 == 0 && ((stbuf.st_mode & S_IFMT) == S_IFREG)))
	    {
	    printf("\n\nThe %s already exists.", obj_desc);

	    opt = 0;
	    while(opt == 0)
		{
		if (strcmp(obj_type, "dbproc") == 0)
		    {
		    printf("\nEnter 'o' - overwrite, 'r' - proceed to rule ");
		    printf("creation, 'q' - quit");
		    }
		else
		    {
		    printf("\nEnter 'o' - overwrite, 'r' - return ");
		    printf(" 'q' - quit");
		    }

		printf("\n  ==> ");

		if (read_line(&opt, 1) == FALSE)
		    opt = 0;

		switch(opt)
		    {
		    case 'o':
			if (remove(ddl_script) != 0)
			    {
			    printf("\nERROR: Cannot overwrite. Run aborts.\n");
			    main_final();
			    exit(1);
			    }

			rc = FALSE;
			rc2 = -1;
			break;

		    case 'r':
			return IGNORE;
			break;

		    case 'q':
			main_final();
			exit(1);
			break;
		
		    default:
			opt = 0;
			printf("\nInvalid response");
			break;
		    }
		}
	    continue;
	    }

	/*
	** DB procedure exists and script file either does not exist or
	** the file name exists, but is not a regular file.
	** If thw DB procedure exists it means the name is used by another
	** table. To get around this rehash the name a little.
	*/

	if (strlen(km_name) >= (size_t) DB_OBJ_LEN)
	    {
	    j = (i/10) + 1;
	    km_name[strlen(km_name) - j] = '\0';
	    }

	sprintf(tmp_str, "%d", i);
	strcat(km_name, tmp_str);
	sprintf(ddl_script, "%s/%s/%s.%s", Default_path, tbl_name,
					   km_name, obj_ext);
	}

    return TRUE;
}


/*
** Use the arguments and a truncation algorithm on the string
** to derive a new name
*/

void get_name(trunc_str, prefix, suffix, max_length, name)
char *trunc_str;
char *prefix;
char *suffix;
int max_length;
char *name;
{
    int len;
    char tmp_string[TMP_LEN];

    if (max_length > TMP_LEN)
	max_length = TMP_LEN;

    if (strlen(trunc_str) > (size_t) TMP_LEN)
	{
	strncpy(tmp_string, trunc_str, TMP_LEN - 1);
	tmp_string[TMP_LEN] = '\0';
	}
    else
	strcpy(tmp_string, trunc_str);

    len = max_length - strlen(prefix) - strlen(suffix);
    trunc_string(tmp_string, len);
    sprintf(name, "%s%s%s", prefix, tmp_string, suffix);

    return;
}


/* Truncate a string to a desired length */

void trunc_string(str, len)
char *str;
int len;
{
    int i, j, l, ignore, no_to_ignore;
    char c;

    l = strlen(str);

    if (l <= len)
        return;

   /* Nice truncate - remove vowels from the rear */

    ignore = 0;
    no_to_ignore = l - len;

    for(i = l - 1; i >= 0; i--)
	{
	c = tolower(str[i]);
	if (c == 'a' || c == 'e' || c == 'i' || c == 'o' || c == 'u')
	    {
	    str[i] = '*';
	    ignore++;

	    if (ignore < no_to_ignore)
		continue;
	    else
		break;
	    }
	}

   /* If string is still too long - remove any underscores from rear */

    if (ignore < no_to_ignore)
	{
	for(i = l - 1; i >= 0; i--)
	    {
	    if (str[i] == '_')
		{
		str[i] = '*';
		ignore++;

		if (ignore < no_to_ignore)
		    continue;
		else
		    break;
		}
	    }
	}

    /* Rebuild the string removing the marked characters */

    for(i = 0, j = 0; str[i] != '\0'; i++)
	{
	if (str[i] == '*')
	    continue;
	else
	    str[j++] = str[i];
	}

    str[j] = '\0';

    /* Do a final truncate (remove from rear) */

    l = strlen(str);
    if (ignore < no_to_ignore)
	str[(l - (no_to_ignore - ignore))] = '\0';

    return;
}


/* Open a file for a DDL SQL script */

void ddl_open(ddl_script, mode)
char *ddl_script;
char *mode;
{
    /* Open script file */

    if ((DDL_ptr = fopen(ddl_script, "w")) == NULL)
        {
        printf("\n\nERROR: Cannot open DDL SQL script");
        printf("\n       file: %s", ddl_script);
        printf("\n       Run aborts.\n");
	main_final();
        exit(1);
        }

    return;
}


/* Append a file to another */

void append_file(o_file, i_file)
char *o_file;
char *i_file;
{
    FILE *o_ptr;
    FILE *i_ptr;

    char buffer[BUFFER_LEN];

    /* Open files */

    if ((o_ptr = fopen(o_file, "a")) == NULL)
        {
        printf("\n\nERROR: Cannot open file - %s", o_file);
        printf("\n       Run aborts.\n");
	main_final();
        exit(1);
        }


    if ((i_ptr = fopen(i_file, "r")) == NULL)
        {
        printf("\n\nERROR: Cannot open file - %s", i_file);
        printf("\n       Run aborts.\n");
	main_final();
        exit(1);
        }

    /* Append */

    while ((fgets(buffer, BUFFER_LEN, i_ptr)) != (char *) NULL)
	{
	if ((fputs(buffer, o_ptr)) == EOF)
	    {
	    printf("\n\nERROR: Cannot append file - %s", i_file);
	    printf("\n       Run aborts.\n");
	    main_final();
	    exit(1);
	    }
	}

    /* Close files */

    fclose(o_ptr);
    fclose(i_ptr);

    return;
}


/* List all the Automatic and Manual related tables */

void list_tables()
{
    int i;
    struct obj_name *tmp_ptr;

    for(i = 0; i < No_auto_relns; i++)
        {
        tmp_ptr = Relns_auto + i;
        fprintf(DDL_ptr, "**\t\t\t%s\n", tmp_ptr->name);
        }

    for(i = 0; i < No_man_relns; i++)
        {
        tmp_ptr = Relns_man + i;
        fprintf(DDL_ptr, "**\t\t\t%s\n", tmp_ptr->name);
        }

    return;
}


/* Add code that checks key columns for default values */

void def_val_code(tbl_str)
char *tbl_str;
{
    int i;
    char def_value[3], opr[4];
    struct key_info *kptr;

    kptr = Keyptr;

    fprintf(DDL_ptr, "\t/* Check for default values */\n\n");

    if (Def_key_vals == 1)
	strcpy(opr, "AND");
    else
	strcpy(opr, "OR ");

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;
	set_def_value(def_value, kptr->c_datatype);

	if (i == 0)
	    fprintf(DDL_ptr, "\tIF  %s = %s\n", kptr->c_name, def_value);
	else
	    fprintf(DDL_ptr, "\t%s %s = %s\n", opr, kptr->c_name, def_value);
	}

    fprintf(DDL_ptr, "\tTHEN\n");
    fprintf(DDL_ptr, "\t\t/* Raise error - key has default values */\n");
    fprintf(DDL_ptr, "\n\t\terror_number = 100004;\n");
    fprintf(DDL_ptr, "\t\tmsg_text = 'Default values may not be used "\
		     "used as the key ';\n");
    fprintf(DDL_ptr, "\t\tmsg_text = msg_text + 'for table: %s.';\n", tbl_str);
    fprintf(DDL_ptr, "\t\tRAISE ERROR error_number msg_text;\n");
    fprintf(DDL_ptr, "\tENDIF;\n\n");

    return;
}


/* Display help information */

void display_help()
{
    printf("\n\nName:   km_express\n");
    printf("Author: Tony Buckley (Waratah Computing Services Pty. Ltd.)\n");
    printf("Usage:  km_express [-v] [-uusername] [-h] dbname\n\n");
    printf("Description:\n");
    printf("km_express is an Ingres utility that will generate database\n");
    printf("rules and procedures to maintain referential integrity\n");
    printf("for defined sets of Tables.\n");
    printf("This is done by checking the system catalogs for possible\n");
    printf("relations to a specified table. The resulting list may be\n");
    printf("amended for additions and deletions.\n");
    printf("It can also generate rules and a procedure to check on\n");
    printf("columns defined as 'not null not default' and raise an error\n");
    printf("if they contain a 0 or '' depending on the datatype.\n\n");
    printf("Environment variables:\n\n");
    printf("USER or LOGNAME - Current user name\n");
    printf("LINES - Number of lines in window / terminal (optional)\n");
    printf("SQL_PATH -\n");
    printf("  Directory path to use to find / create script files.\n");
    printf("  Script files are created in directories below this path\n");
    printf("  with directories being named after the Table in use.\n");
    printf("  For example, you might set SQL_PATH to '/usr/dba/ddl' and\n");
    printf("  Table name directories will be created below this as needed.\n");
    printf("  If SQL_PATH is not set, the current directory is used.\n");
    printf("DEF_KVALS -\n");
    printf("  1 - Some, but not all key columns may have a default value\n");
    printf("      (0 or '') depending on the datatype. (Default).\n");
    printf("  2 - No key column may have a default value.\n");
    printf("  3 - Don't care about default values.\n\n");

    return;
}
