/*
** This code is the property of Waratah Computing Services - 22/10/93.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions: 	General Ingres installation routines
**
** Description: Routines to check Ingres installation parameters.
**
*/

/* Headers */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define INGPRENV1 "$II_SYSTEM/ingres/bin/ingprenv1 "

/* Function Prototypes */

int getingenv(char *, char *);
void set_type_length(int, char *, char *);
void set_def_value(char *, char *);



/* Global Variables */



/* FUNCTION CODE */


/* Find and return an Ingres environment variable */

int getingenv(name, value)
char *name;
char *value;
{
    FILE *pipe;
    char cmd[100], buffer[1024];
    char *p;

    sprintf(cmd, "%s%s", INGPRENV1, name);

    if ((pipe = popen(cmd, "r")) == NULL)
	{
	value[0] = '\0';
	return FALSE;
	}

    /* Read Ingres environment variable */

    if ((fgets(buffer, 1023, pipe)) != (char *) NULL)
        {
        if ((p = strchr(buffer, '\n')) != (char *) NULL)
	    *p = '\0';

	strcpy(value, buffer);
        }
    else
	value[0] = '\0';

    /* Clean up */

    if (pclose(pipe) == 0)
	return TRUE;
    else
	return FALSE;
}


/* Arrange the length and datatype as required */

void set_type_length(length, datatype, desc_str)
int length;
char *datatype;
char *desc_str;
{
    char c;

    /* Money and Date data types */

    if (length == 0)
	{
	strcpy(desc_str, datatype);
	return;
	}

    /* Define the other datatypes */

    c = datatype[0];
    switch(c)
	{
	/* c, char (character), text and varchar */

	case 'c': case 't': case 'v':
	    sprintf(desc_str, "%s(%d)", datatype, length);
	    break;

	/* integer and float */

	case 'i': case 'f':
	    sprintf(desc_str, "%s%d", datatype, length);
	    break;

	/* smallint */

	case 's':
	    strcpy(desc_str, "integer2");
	    break;

	/* real and double precision */

	case 'r': case 'd':
	    strcpy(desc_str, "float8");
	    break;

	/* unknown */

	default:
	    printf("\nWARNING: Unknown datatype - %s\n", datatype);
	    strcpy(desc_str, datatype);
	    break;
	}

    return;
}


/* Determine the default, empty or no value for a particular datatype */

void set_def_value(def_value, datatype)
char *def_value;
char *datatype;
{
    char c;

    /* Define the datatypes */

    c = datatype[0];
    switch(c)
	{
	/* date and double precision */

	case 'd':
	    if (datatype[1] == 'a')
		strcpy(def_value, "''");
	    else
		strcpy(def_value, "0");

	/* c, char (character), text and varchar */

	case 'c': case 't': case 'v':
	    strcpy(def_value, "''");
	    break;

	/* smallint, integer, money, float and real */

	case 'i': case 'f': case 's': case 'm': case 'r':
	    strcpy(def_value, "0");
	    break;

	/* unknown */

	default:
	    printf("\nWARNING: Unknown datatype - %s\n", datatype);
	    strcpy(def_value, "''");
	    break;
	}

    return;
}
