/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
**
** Program:	km_express
**
** Description: Program to generate Knowledge Management extension rules and
**		database procedures to maintain referential integrity.
**
** Arguments:  	Username (Optional)
**		Verbose
**		Help
**		Database Name
**
** Usage:	km_express [-v] [-uusername] [-h] dbname
**
** Environment:	USER or LOGNAME - Current user name
**		LINES - Number of lines in window / terminal (optional)
**		SQL_PATH - Directory path to use to find Table script
**			   directories (eg. /usr/dba/ddl/table_name).
**			   If not present, current directory is used and
**			   table directories are created if not present.
**		DEF_KVALS - Check the values of key columns as follows:-
**			     1 - Some, but not all key columns may have a
**				 default value (0 or '') depending on the
**				 datatype. (Default).
**			     2 - No key column may have a default value.
**			     3 - Don't care about default values.
*/

/* Headers */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>


/* Macro Definitions */

#define TRUE 1
#define FALSE 0
#define TEMPLATE "km_express [-uusername] [-h] database"
#define DB_OBJ_LEN 24
#define DB_OBJ_LEN_S 25
#define MAN_RELS 50
#define TABLES 100
#define FILE_LEN 256
#define BUFFER_LEN 100
#define DEF_LINES 25
#define DELIM '|'
#define MISC "misc_dbprocs"
#define VERSION "(sql/1.0)"


/* Function Prototypes */

int main(int, char*[]);
void main_init(int, char*[]);
void check_args(int, char*[]);
void run_date();
void clear_screen();
void main_final();
void km_menu();
void display_menu();
void check_superuser(char *, char *);
void get_table();
int get_table_2();
int get_columns(char *);
void get_relations();
void relation_mtce();
void table_list_del();
void table_list_add();
int read_line(char *, int);
void casc_upd_generate();
void ref_generate();
void res_del_generate();
void casc_del_generate();
void no_default_generate();
int define_rels();
void dir_find(char *);
void match_keys(char *);

extern int sql_lookup(char*, char*);
extern int sql_related();
extern int cascade_upd_dbproc();
extern void cascade_upd_rule();
extern int ref_dbproc();
extern void ref_rule(char *, int);
extern int restrict_del_dbproc();
extern void restrict_del_rule();
extern int cascade_del_dbproc();
extern void cascade_del_rule();
extern int no_default_dbproc();
extern void no_default_rule(char *, char *, char *);
extern void lookup_keys();
extern void lookup_cols(char *);
extern int save_keys();
extern int save_cols();
extern int getingenv(char *, char *);
extern void display_help();


/* Structures */

struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

struct col_info
{
char c_name[33];
char c_datatype[33];
};

struct obj_name
{
char name[33];
};


/* Global Variables */

FILE *I_PTR, *O_PTR;

int Display_lines, Def_key_vals;
int No_key_cols, No_auto_relns, No_man_relns, No_tables, No_columns;
int Verbose, No_def_dbproc, Create_std;

char Dbname[33];
char Table_name[33], Dbproc_name[DB_OBJ_LEN_S];
char Rdate[25], Username[25];
char Default_path[FILE_LEN];
char Restriction[500];
char Log_file[FILE_LEN];

char *Lines, *User, *Path, *Sql_in, *Sql_out, *Env;

struct obj_name *Relns_auto, *Relns_man, *Man_cols, *Tables;
struct key_info *Keyptr;
struct col_info *Colptr;

extern struct obj_name *Old_names;



/* FUNCTION CODE */

int main(argc, argv)
int argc;
char *argv[];
{
    main_init(argc, argv);

    km_menu();

    main_final();

    exit(0);
}


/* Initial set up */

void main_init(argc, argv)
int argc;
char *argv[];
{
    int rc;
    char tmp_dir[FILE_LEN];

    printf("\n\nKnowledge Management generation starting ...");

    /* Argument check */

    if (argc < 2 || argc > 5)
        {
        printf("\n\nERROR: Invalid arguments supplied");
        printf("\n       Usage: %s\n", TEMPLATE);
        exit(1);
        }

    /* Get temporary work file names */

    rc = getingenv("II_TEMPORARY", tmp_dir);
    if (rc == FALSE || tmp_dir[0] == '\0')
	{
	printf("\n\nERROR: Unable to read ingres environment variable - ");
	printf("II_TEMPORARY.");
	printf("\n       See your database administrator.\n");
	exit(1);
	}

    Sql_in = tempnam(tmp_dir, "sql_");
    Sql_out = tempnam(tmp_dir, "sql_");

    /* Get any environment variables */

    Def_key_vals = 1;

    if ((Env = getenv("DEF_KVALS")) != (char *) NULL)
	{
	Def_key_vals = atoi(Env);
	if (Def_key_vals < 0 || Def_key_vals > 3)
	    {
	    printf("\n\nERROR: Invalid value for Env. variable - DEF_KVALS\n");
	    main_final();
	    exit(1);
	    }
	}

    if ((User = getenv("USER")) == (char *) NULL)
	if ((User = getenv("LOGNAME")) == (char *) NULL)
	    {
	    printf("\n\nERROR: Unable to get USER or LOGNAME environment ");
	    printf("variable");
	    printf("\n       Set up either of these and set it to your");
	    printf("\n       login id and try again\n");
	    main_final();
	    exit(1);
	    }

    strcpy(Username, User);

    if ((Lines = getenv("LINES")) == (char *) NULL)
	Display_lines = DEF_LINES;
    else
	Display_lines = atoi(Lines);

    if ((Path = getenv("SQL_PATH")) == (char *) NULL)
	strcpy(Default_path, ".");
    else
	strcpy(Default_path, Path);

    /* Validate User */

    sprintf(Restriction, "name = '%s'", Username);
    rc = sql_lookup("iidbdb", "iiuser");

    if (rc == FALSE)
	{
	printf("\n\nERROR: You are not a valid Ingres user\n");
	main_final();
        exit(1);
	}

    /* Set up log file */

    sprintf(Log_file, "%s_ddl.log", Username);

    /* Validate command line arguments */

    check_args(argc, argv);

    /* Validate database */

    sprintf(Restriction, "name = '%s'", Dbname);
    rc = sql_lookup("iidbdb", "iidatabase");

    if (rc == FALSE)
	{
	printf("\n\nERROR: The database - %s - does not exist", argv[1]);
        printf("\n       Usage: %s\n", TEMPLATE);
	main_final();
        exit(1);
	}

    /* Run date */

    run_date();

    /* Any field initialisations */

    No_def_dbproc = FALSE;

    return;
}


/* Check for -u flag - validate current user has authority */

void check_args(argc, argv)
int argc;
char *argv[];
{
    int i, c;
    char *p;

    Dbname[0] = '\0';
    Verbose = FALSE;

    for(i = 1; i < argc; i++)
	{
	p = argv[i];
	c = *p;

	switch(c)
	    {
	    case '-':
		c = *(++p);
		switch(c)
		    {
		    case 'u':
			p++;
			check_superuser(Username, p);
			strcpy(Username, p);
			break;

		    case 'v':
			Verbose = TRUE;
			break;

		    case 'h':
			display_help();
			main_final();
			exit(1);
			break;

		    default:
			printf("\n\nERROR: Invalid argument: -%c", c);
			printf("\n       Usage: %s\n", TEMPLATE);
			main_final();
			exit(1);
		    }
		break;

	    default:
		strcpy(Dbname, p);
		break;
	    }
	}

    if (Dbname[0] == '\0')
	{
	printf("\n\nERROR: Database name not supplied");
	printf("\n       Usage: %s\n", TEMPLATE);
	main_final();
	exit(1);
	}

    return;
}


/* Check for superuser authority */

void check_superuser(user, superuser)
char *user;
char *superuser;
{
    int rc;

    sprintf(Restriction, "name = '%s' and status > 25", user);
    rc = sql_lookup("iidbdb", "iiuser");

    if (rc == FALSE)
	{
	printf("\n\nERROR: You do not have permission to use the -u flag\n");
	main_final();
        exit(1);
	}

    sprintf(Restriction, "name = '%s'", superuser);
    rc = sql_lookup("iidbdb", "iiuser");

    if (rc == FALSE)
	{
	printf("\n\nERROR: %s is not a valid Ingres user\n", superuser);
	main_final();
        exit(1);
	}

    return;
}


/* Get today's date */

void run_date()
{
    struct tm sys_time;
    struct tm *p_sys = &sys_time;
    time_t cal_time;

    time(&cal_time);
    p_sys = localtime(&cal_time);
    strftime(Rdate, 25, "%d/%m/%Y %H:%M", p_sys);

    return;
}


/* Finalise */

void main_final()
{
    clear_screen();

    if (remove(Sql_in) != 0)
	printf("\n\nERROR: Could not remove work file: %s", Sql_in);

    if (remove(Sql_out) != 0)
	printf("\n\nERROR: Could not remove work file: %s", Sql_out);

    printf("\n\nA log of this session's SQL (if any) is in %s", Log_file);
    printf("\nKnowledge Management generation completed\n");

    return;
}


/* Clear the display */

void clear_screen()
{
    int i;

    for(i = 0; i < Display_lines; i++)
	printf("\n");

    return;
}


/* Main menu and control processing */

void km_menu()
{
    char opt;

    opt = 0;
    while (opt != '7' && tolower(opt) != 'q')
	{
	Create_std = FALSE;
	clear_screen();
	display_menu();
	if (read_line(&opt, 1) == FALSE)
	    opt = 0;

	switch(opt)
	    {
	    /* Default action - 2, 3 and 4 */

	    case '1':
		casc_upd_generate();

		if (Create_std == -1)
		    break;

		Create_std = TRUE;
		ref_generate();
		res_del_generate();
		free(Relns_auto);
		free(Relns_man);
		free(Man_cols);
		free(Keyptr);
		free(Old_names);
		break;

	    /* Insert rule */

	    case '2':
		ref_generate();
		free(Relns_auto);
		free(Relns_man);
		free(Man_cols);
		free(Keyptr);
		break;

	    /* Update rule */

	    case '3':
		casc_upd_generate();
		free(Relns_auto);
		free(Relns_man);
		free(Man_cols);
		free(Keyptr);
		free(Old_names);
		break;

	    /* Restricted delete rule */

	    case '4':
		res_del_generate();
		free(Relns_auto);
		free(Relns_man);
		free(Man_cols);
		free(Keyptr);
		break;

	    /* Cascade delete rule */

	    case '5':
		casc_del_generate();
		free(Relns_auto);
		free(Relns_man);
		free(Man_cols);
		free(Keyptr);
		break;

	    /* Mandatory Column rules */

	    case '6':
		no_default_generate();
		free(Tables);
		break;

	    /* Quit */

	    case '7': case 'q':
		break;

	    default:
		printf("\nInvalid Option\n");
		break;
	    }
	}

    return;
}


/* Display available options */

void display_menu()
{
    printf("\nWaratah Computing Services Pty. Ltd. (1995)  %s", VERSION);
    printf("\n\n\t    K N O W L E D G E   M A N A G E M E N T   E X P R E S S");
    printf("\n\n\t\t\tReferential Integrity Generation");
    printf("\n\t\t\t----------- --------- ----------");
    printf("\n\n\n\t\t1.  Default Integrity (2, 3 and 4)  (recommended)");
    printf("\n\n\t\t2.  Reference on Insert / Update");
    printf("\n\n\t\t3.  Cascade on Update");
    printf("\n\n\t\t4.  Restrict on Delete");
    printf("\n\n\t\t5.  Cascade on Delete");
    printf("\n\n\t\t6.  Mandatory Columns");
    printf("\n\n\t\t7.  Quit");
    printf("\n\n\t\t    Please enter your selection ==> ");

    return;
}


/* Prompt the user for the 'Parent' relation */

void get_table()
{
    int rc;

    clear_screen();

    Table_name[0] = '\0';
    while (Table_name[0] == '\0')
	{
	printf("\n\nYou need to enter the name of a Table that starts ");
	printf("a relation\n(ie a 'Parent' table).");
	printf("\n\n\tTable name ==> ");

	if (read_line(Table_name, DB_OBJ_LEN) == FALSE)
	    {
	    printf("\n\nERROR: Table names have a maximum of %d characters",
							DB_OBJ_LEN);
	    printf("\n       Try again or Control & C to quit");
	    Table_name[0] = '\0';
	    continue;
	    }

	/* Check if valid table */

	sprintf(Restriction, "table_name = '%s'\nand table_owner = '%s'",
			Table_name, Username);
	rc = sql_lookup(Dbname, "iitables");

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Table - %s - does not exist", Table_name);
	    Table_name[0] = '\0';
	    continue;
	    }
	}

    /* Get table key information */

    lookup_keys();
    rc = save_keys();
    if (rc == FALSE)
	{
	printf("\n\nERROR: Table - %s - does not have a Primary key.",
						Table_name);
	printf("\n       You will have to do this table's KM manually");
	printf("\n       or give it a Primary key (even temporary one\n");
	main_final();
	exit(1);
	}

    return;
}


/* Build a list of possible related tables and edit if necessary */

void get_relations()
{
    int rc;

    printf("\n\nThe system catalogs will now be examined to build a ");
    printf("\nlist of tables that may relate to table - %s.", Table_name);
    printf("\nYou will be able to edit this list (Delete and Add).");
    printf("\n\nLooking for (possible) related tables ...");

    /* Find related tables - an array of strings (tables) is derived */

    No_auto_relns = No_man_relns = 0;
    rc = sql_related();

    if (rc == FALSE)
	{
	printf("\n\nERROR: Could not read (possible) related tables");
	printf("\n       See file - %s - for more details\n", Sql_out);
	exit(1);
	}

    relation_mtce();

    return;
}


/* Control the maintenance of the list of related tables */

void relation_mtce()
{
    int i;
    char buffer[BUFFER_LEN];
    char *p;

    /* Open the file */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
        {
        printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
        exit(1);
        }

    /* Count the number of (possible) tables */

    i = 0;
    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "OBJ_FOUND")) != (char *) NULL)
	    i++;
	}

    fclose(O_PTR);

    /* Allocate the maximum space */

    if (i == 0)
	i = 1;

    Relns_auto = (struct obj_name *) calloc(i, sizeof(struct obj_name));
    if (Relns_auto  == NULL)
	{
	printf("\nERROR: No space available for list of related tables\n");
	main_final();
	exit(2);
	}

    /* Open the file again */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
        {
        printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
        exit(1);
        }

    /* Edit the list for deletions and additions */

    table_list_del();
    table_list_add();

    fclose(O_PTR);

    return;
}


/* Browse the list of (possibly) related tables and confirm or delete */

void table_list_del()
{
    int i, j, found = FALSE;
    char c;
    char tmp_table[DB_OBJ_LEN_S], buffer[BUFFER_LEN];
    char *p;
    struct obj_name *tmp_ptr;

    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "OBJ_FOUND")) == (char *) NULL)
	    continue;

	if (found == FALSE)
	    {
	    clear_screen();
	    printf("\n\nThe following tables may relate to - %s", Table_name);
	    printf("\nPlease Confirm or Delete\n");
	    found = TRUE;
	    }

	for(i = 1, j = 0; buffer[i] != ' ' && buffer[i] != '\0'; i++, j++)
	    {
	    if (buffer[i] == DELIM)
		break;
	    tmp_table[j] = buffer[i];
	    }
	tmp_table[j] = '\0';

	/* Check for error */

	if (buffer[i] == '\0')
	    {
	    printf("\n\nERROR: Failed to read (possible) related tables");
	    printf("\n       See the file - %s - for details\n", Sql_out);
	    exit(1);
	    }

	while (1)
	    {
	    printf("\n\t%s\t Enter (c)onfirm or (d)elete ==> ", tmp_table);

	    if (read_line(&c, 1) == FALSE)
		c = FALSE;

	    if (c == 'c' || c == 'd')
		break;

	    printf("\nInvalid response - please enter (c)onfirm or (d)elete");
	    }

	if (c == 'c')
	    {
	    tmp_ptr = Relns_auto + No_auto_relns;
	    strcpy(tmp_ptr->name, tmp_table);
	    No_auto_relns++;
	    }
	}

    if (found == FALSE)
	printf("\n\nNo (possibly) related tables were found.");

    return;
}


/* Allow additional relations - need to match new column names now */

void table_list_add()
{
    int rc, alloc_size, ins_count = 0;
    char tmp_table[DB_OBJ_LEN_S];
    struct obj_name *tmp_ptr;

    printf("\n\n\nYou may add related tables to the previous list (if any).");
    printf("\nEnter table names and key columns when prompted and press ENTER");
    printf("\nEnter Control & D when finished\n");

    /* Allocate some initial space for manual tables and their columns */

    Relns_man = (struct obj_name *) calloc(MAN_RELS, sizeof(struct obj_name));
    if (Relns_man == NULL)
	{
	printf("\nERROR: No space available for list of related tables\n");
	main_final();
	exit(2);
	}

    alloc_size = MAN_RELS * No_key_cols;
    Man_cols = (struct obj_name *) calloc(alloc_size, sizeof(struct obj_name));
    if (Man_cols == NULL)
	{
	printf("\nERROR: No space available for related table columns\n");
	main_final();
	exit(2);
	}

    while (1)
	{
	printf("\n Table ==> ");

	rc = read_line(tmp_table, DB_OBJ_LEN);

	if (rc == EOF)
	    break;

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Table names have a maximum of %d characters",
							DB_OBJ_LEN);
	    printf("\n       Try again or Control & C to quit");
	    continue;
	    }

	/* Check if valid table */

	sprintf(Restriction, "table_name = '%s'\nand table_owner = '%s'",
			tmp_table, Username);
	rc = sql_lookup(Dbname, "iitables");

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Table - %s - does not exist", tmp_table);
	    printf("\n       Try again or Control & C to quit");
	    continue;
	    }

	ins_count++;

	/* Get more space if needed */

	if (ins_count > MAN_RELS)
	    {
	    printf("\nWARNINNG: The number of tables that you may add at");
	    printf("\n          this time is limited to %d.", MAN_RELS);
	    printf("\n          You will need to do another run to add any");
	    printf("\n          more. Table %s not included.\n", tmp_table);
	    break;
	    }

	tmp_ptr = Relns_man + No_man_relns;
	strcpy(tmp_ptr->name, tmp_table);
	No_man_relns++;

	/* Now add key columns */

	match_keys(tmp_table);
	}

    return;
}


/* These tables will not have matching key column names - do manually */

void match_keys(table)
char *table;
{
    int i, rc;
    char tmp_col[DB_OBJ_LEN_S];
    struct key_info *kptr;
    struct obj_name *tmp_ptr;

    for(i = 0; i < No_key_cols;)
	{
	kptr = Keyptr + i;

	printf("\n\tPlease enter the foreign key column name that relates to");
	printf("\n\tprimary key column - %s - in %s", kptr->c_name,Table_name);
	printf("\n\t Column ==> ");

	rc = read_line(tmp_col, DB_OBJ_LEN);

	if (rc == EOF)
	

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Column names have a maximum of %d characters",
							DB_OBJ_LEN);
	    printf("\n       Try again or Control & C to quit");
	    continue;
	    }

	/* Check if valid column */

	sprintf(Restriction, "table_name = '%s'\n"\
			     "and table_owner = '%s'"\
			     "and column_name = '%s'",
			     table, Username, tmp_col);
	rc = sql_lookup(Dbname, "iicolumns");

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Column - %s - does not exist", tmp_col);
	    printf("\n       Try again or Control & C to quit");
	    continue;
	    }

	tmp_ptr = Man_cols + i;
	strcpy(tmp_ptr->name, tmp_col);
	i++;
	}

    return;
}


/* Prompt the user for a list of tables for no default value domain rules */

int get_table_2()
{
    int rc;
    struct obj_name *tmp_ptr;

    /* Allocate some initial space for tables */

    Tables = (struct obj_name *) calloc(TABLES, sizeof(struct obj_name));
    if (Tables == NULL)
	{
	printf("\nERROR: No space available for list of tables\n");
	main_final();
	exit(2);
	}

    clear_screen();
    Table_name[0] = '\0';
    No_tables = 0;

    printf("\nYou need to enter the name(s) of Table(s) that you wish");
    printf("\nto create no default value (ie. mandatory column) "\
	   "domain rules for.");
    printf("\nA rule will be created for all 'not null / not default'");
    printf(" columns.");
    printf("\nEnter a table name at the prompt and press ENTER");
    printf("\n\nEnter Control & D when finished\n");

    while (1)
	{
	printf("\n\tTable name ==> ");

	rc = read_line(Table_name, DB_OBJ_LEN);

	if (rc == EOF)
	    break;

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Table names have a maximum of %d characters",
							DB_OBJ_LEN);
	    printf("\n       Try again or Control & C to quit");
	    Table_name[0] = '\0';
	    continue;
	    }

	/* Check if valid table */

	sprintf(Restriction, "table_name = '%s'\nand table_owner = '%s'",
			Table_name, Username);
	rc = sql_lookup(Dbname, "iitables");

	if (rc == FALSE)
	    {
	    printf("\n\nERROR: Table - %s - does not exist", Table_name);
	    Table_name[0] = '\0';
	    continue;
	    }

	/* Get more space if needed */

	if ((No_tables + 1) >= TABLES)
	    {
	    printf("\nWARNINNG: The number of tables that you may add at");
	    printf("\n          this time is limited to %d.", TABLES);
	    printf("\n          You will need to do another run to add any");
	    printf("\n          more. Table %s not included.\n", Table_name);
	    break;
	    }

	tmp_ptr = Tables + No_tables;
	strcpy(tmp_ptr->name, Table_name);
	No_tables++;
	}

    if (No_tables == 0)
	return FALSE;
    else
	return TRUE;
}


/* Retrieve & store all mandatory column details for a table */

int get_columns(table)
char *table;
{
    lookup_cols(table);

    if (save_cols() == FALSE)
	{
	printf("\n\nWARNING: Table - %s - has no mandatory columns.", table);
	printf("\n         This should be investigated.\n");
	return FALSE;
	}

    return TRUE;
}


/* Read a specified maximum number of characters and return */

int read_line(p, n)
char *p;
int n;
{
    int i, c;

    for (i = 0, *p = '\0'; (c = getchar()) != EOF; i++)
	{
	if (c == '\n')
	    {
	    i--;
	    break;
	    }

	if (i < n)
	    *(p + i) = c;
	}

    if (c == EOF)
	return EOF;

    if (i >= n)
	return FALSE;

    if (i > 1)
	*(p + i + 1) = '\0';

    return TRUE;
}


/* Control processing to generate cascade update rule and procedure */

void casc_upd_generate()
{
    int rc;

    if (define_rels() == FALSE)
	{
	Create_std = -1;
	return;
	}

    /* Construct the KM code */

    printf("\n\nCascade update process starting ...");

    rc = cascade_upd_dbproc();

    if (rc == TRUE)
	cascade_upd_rule();
    else
	{
	printf("\n\nRule generation process not continued due to error(s) in");
	printf("\nDatabase Procedure processing.");
	}

    return;
}


/* Control processing to generate reference rule(s) and procedure */

void ref_generate()
{
    int i, rc;
    struct obj_name *tmp_ptr;

    /* Construct the KM code */

    if (Create_std == FALSE)
	if (define_rels() == FALSE)
	    return;

    printf("\n\nReference insert / update process starting ...");

    rc = ref_dbproc();

    /* Reference rules will reside in the 'child' table directory */

    if (rc == TRUE)
	{
	for(i = 0; i < No_auto_relns; i++)
	    {
	    tmp_ptr = Relns_auto + i;
	    dir_find(tmp_ptr->name);
	    ref_rule(tmp_ptr->name, -1);
	    }

	for(i = 0; i < No_man_relns; i++)
	    {
	    tmp_ptr = Relns_man + i;
	    dir_find(tmp_ptr->name);
	    ref_rule(tmp_ptr->name, i * No_key_cols);
	    }
	}
    else
	{
	printf("\n\nRule generation process not continued due to error(s) in");
	printf("\nDatabase Procedure processing.");
	}

    return;
}


/* Control processing to generate restricted delete rule and procedure */

void res_del_generate()
{
    int rc;

    if (Create_std == FALSE)
	if (define_rels() == FALSE)
	    return;

    /* Construct the KM code */

    printf("\n\nRestricted delete process starting ...");

    rc = restrict_del_dbproc();

    if (rc == TRUE)
	restrict_del_rule();
    else
	{
	printf("\n\nRule generation process not continued due to error(s) in");
	printf("\nDatabase Procedure processing.");
	}

    return;
}


/* Control processing to generate cascade delete rule and procedure */

void casc_del_generate()
{
    int rc;

    if (define_rels() == FALSE)
	return;

    /* Construct the KM code */

    printf("\n\nCascade delete process starting ...");

    rc = cascade_del_dbproc();

    if (rc == TRUE)
	cascade_del_rule();
    else
	{
	printf("\n\nRule generation process not continued due to error(s) in");
	printf("\nDatabase Procedure processing.");
	}

    return;
}


/* Control processing to generate no default value domain rule and procedure */

void no_default_generate()
{
    int i, j, rc;
    struct col_info *cptr;
    struct obj_name *tmp_ptr;


    if (get_table_2() == FALSE)
	return;

    /* Construct the KM code */

    if (No_def_dbproc == TRUE)
	rc = TRUE;
    else
	{
	dir_find(MISC);

	printf("\n\nMandatory column process starting ...");

	rc = no_default_dbproc();
	}

    if (rc == TRUE)
	{
	No_def_dbproc = TRUE;

	for(i = 0; i < No_tables; i++)
	    {
	    tmp_ptr = Tables + i;
	    dir_find(tmp_ptr->name);

	    if (get_columns(tmp_ptr->name) == TRUE)
		{
		printf("\n\nProcessing table - %s ...\n", tmp_ptr->name);

		for(j = 0; j < No_columns; j++)
		    {
		    cptr = Colptr + j;
		    no_default_rule(tmp_ptr->name, cptr->c_name,
				    cptr->c_datatype);
		    }

		free(Colptr);
		}
	    }
	}
    else
	{
	printf("\n\nRule generation process not continued due to error(s) in");
	printf("\nDatabase Procedure processing.");
	}

    return;
}


/* Define which relationships to process */

int define_rels()
{
    char c;

    /* Get the relation owner & check directory */

    get_table();
    dir_find(Table_name);

    /* What tables relate to this table */

    get_relations();

    if (No_auto_relns == 0 && No_man_relns == 0)
	{
	printf("\n\nWARNING: You have not defined any relationships for ");
	printf("Table - %s", Table_name);
	printf("\nPress ENTER to continue:");
	read_line(&c, 1);
	return FALSE;
	}

    return TRUE;
}


/* See if a directory exists and create if necessary */

void dir_find(tbl_name)
char *tbl_name;
{
    int rc;
    char dir_name[FILE_LEN], sys_cmd[100];
    struct stat stbuf;

    sprintf(dir_name, "%s/%s", Default_path, tbl_name);
    rc = stat(dir_name, &stbuf);

    if (rc == 0 && ((stbuf.st_mode & S_IFMT) != S_IFDIR))
	{
	printf("\n\nERROR: There is a file entry for - %s -,", Table_name);
	printf("\n       but it is not a directory. There must either be");
	printf("\n       a directory present for each table or one must be");
	printf("\n       able to be created by the program\n");
	main_final();
	exit(1);
	}

    if (rc == -1)
	{
	sprintf(sys_cmd, "mkdir %s", dir_name);
	system(sys_cmd);
	}

    return;
}
