/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions:	Routines to generate Knowledge Management rule and
**		procedure for columns that must have a value other than
**		the default value.
*/

/* Headers */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef IGNORE
#define IGNORE 9
#endif

#ifndef DB_OBJ_LEN
#define DB_OBJ_LEN 24
#endif

#ifndef DB_OBJ_LEN_S
#define DB_OBJ_LEN_S 25
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif

#ifndef MISC
#define MISC "misc_dbprocs"
#endif

#ifndef MAND_COL
#define MAND_COL "mandatory_col"
#endif



/* Function Prototypes */

int no_default_dbproc();
int mand_col_dbproc(char *);
void mand_col_sql();
void final_no_def();
void no_default_rule(char *, char *, char *);
int mand_col_rule(char *, char *, char *, char *, char *);
void mand_col_rule_sql(char *, char *, char *, char *);

extern int check_existing(char *, char *, char *, char *);
extern void get_name(char *, char *, char *, int, char *);
extern int run_ddl(char *);
extern void set_def_value(char *, char *);


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};


/* Global Variables */

FILE *DDL_ptr;

extern char Dbproc_name[DB_OBJ_LEN_S];
extern char Rdate[25], Username[25];



/* FUNCTION CODE */


/*
** Generate a no default value database procedure, if necessary.
** The main logic of this INSERT & UPDATE function is that if a
** Column in a Table is defined as 'not null / not default' then
** the value entered for the column must not be a default.
** IE. It is a mandatory column.
*/

int no_default_dbproc()
{
    int result;
    char ddl_script[FILE_LEN];

    strcpy(Dbproc_name, MAND_COL);
    if (check_existing(MISC, Dbproc_name, ddl_script, "dbproc") == IGNORE)
	return TRUE;

    printf("\n\nCreating database procedure for Mandatory Column ...");
    result = mand_col_dbproc(ddl_script);
    final_no_def(result);

    return result;
}


/* Finalise */

void final_no_def(result)
int result;
{
    if (result == TRUE)
	printf("\nCreated ok\n");
    else
	printf("\nDDL script created, but it contains errors\n");

    system("sleep 2");

    return;
}


/* Control creation of ddl sql script file for a DB Procedure */

int mand_col_dbproc(ddl_script)
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    mand_col_sql();
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare database procedure */

void mand_col_sql()
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Procedure:  %s\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.dbproc\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database Procedure to raise an error when a "\
		     "column that must have a\n");
    fprintf(DDL_ptr, "**   value gets a default value.\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop procedure before creating  */\n\n");
    fprintf(DDL_ptr, "DROP PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create procedure %s  */\n\n", Dbproc_name);
    fprintf(DDL_ptr, "CREATE PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "(\n");

    /* Arguments */

    fprintf(DDL_ptr, "    table_name\t\tvarchar(32)\tnot null not default,\n");
    fprintf(DDL_ptr, "    column_name\t\tvarchar(32)\tnot null not default\n");
    fprintf(DDL_ptr, ") =\n");

    /* Variables */

    fprintf(DDL_ptr, "DECLARE\n");
    fprintf(DDL_ptr, "\terror_number =\tinteger4\tnot null with default;\n");
    fprintf(DDL_ptr, "\tmsg_text =\tvarchar(240)\tnot null with default;\n\n");

    /* Main logic */

    fprintf(DDL_ptr, "BEGIN\n");
    fprintf(DDL_ptr, "\t/* Raise error - value was not entered */\n");
    fprintf(DDL_ptr, "\n\terror_number = 100003;\n");
    fprintf(DDL_ptr, "\tmsg_text = 'Please enter a value for ' + column_name"\
		     ";\n");
    fprintf(DDL_ptr, "\tmsg_text = msg_text + ' in ' + table_name + '.';\n");
    fprintf(DDL_ptr, "\tRAISE ERROR error_number msg_text;\n");
    fprintf(DDL_ptr, "END;\n\\p\\g\n");

    return;
}


/*
** Generate a mandatory column database rule, if necessary.
** See no default value database procedure comments for details.
*/

void no_default_rule(table, column, datatype)
char *table;
char *column;
char *datatype;
{
    int result;
    char rule_name[DB_OBJ_LEN_S];
    char ddl_script[FILE_LEN];
    char prefix[5];

    strcpy(prefix, table);
    trunc_string(prefix, 3);
    strcat(prefix, "_");

    get_name(column, prefix, "_mand", DB_OBJ_LEN, rule_name);
    if (check_existing(table, rule_name, ddl_script, "rule") == IGNORE)
	return;

    printf("\n\nCreating database rule for Mandatory Column (%s) ...", column);
    result = mand_col_rule(rule_name, ddl_script, table, column, datatype);
    final_no_def(result);

    return;
}


/* Control creation of ddl sql script file for a DB Rule */

int mand_col_rule(rule_name, ddl_script, table, column, datatype)
char *rule_name;
char *ddl_script;
char *table;
char *column;
char *datatype;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    mand_col_rule_sql(rule_name, table, column, datatype);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare database rule */

void mand_col_rule_sql(rule_name, table, column, datatype)
char *rule_name;
char *table;
char *column;
char *datatype;
{
    char def_value[3];

    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Rule:  %s\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.rule\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database rule to implement domain check "\
		     "for a column that must\n");
    fprintf(DDL_ptr, "**   not have a default value.\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**\t\tTable  - %s\n", table);
    fprintf(DDL_ptr, "**\t\tColumn - %s\n", column);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop rule before creating  */\n\n");
    fprintf(DDL_ptr, "DROP RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create Rule %s  */\n\n", rule_name);
    fprintf(DDL_ptr, "CREATE RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\n");

    /* Rule action */

    fprintf(DDL_ptr, "\tAFTER INSERT, UPDATE (%s) ON %s\n", column, table);
    set_def_value(def_value, datatype);
    fprintf(DDL_ptr, "\t\tWHERE new.%s = %s\n\n", column, def_value);
    fprintf(DDL_ptr, "\tEXECUTE PROCEDURE %s(\n", Dbproc_name);

    /* Rule arguments */

    fprintf(DDL_ptr, "\t\t\ttable_name = '%s',\n", table);
    fprintf(DDL_ptr, "\t\t\tcolumn_name = '%s')\n", column);

    fprintf(DDL_ptr, "\\p\\g\n");
    fclose(DDL_ptr);

    return;
}
