/*
** This code is the property of Waratah Computing Services - 28/05/95.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions:	Routines to generate Knowledge Management rule(s) and
**		procedure for reference insert and update.
**
*/

/* Headers */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef IGNORE
#define IGNORE 9
#endif

#ifndef DB_OBJ_LEN
#define DB_OBJ_LEN 24
#endif

#ifndef DB_OBJ_LEN_S
#define DB_OBJ_LEN_S 25
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif



/* Function Prototypes */

int ref_dbproc();
int ref_ins_upd_dbproc(char *);
int ref_ins_upd_rule(char *, char *, char *, int);
void ref_dbp_header();
void ref_dbproc_args();
void ref_dbproc_sql();
void final_ref();
void ref_rule(char *, int);
void ref_rule_header(char *, char *);
void ref_rule_action(char *);
void ref_rule_args(char *, int);

extern int check_existing(char *, char *, char *, char *);
extern void get_name(char *, char *, char *, int, char *);
extern int run_ddl(char *);
extern void set_def_value(char *, char *);
extern void list_tables();
extern void def_val_code(char *);


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

extern struct obj_name
{
char name[33];
};


/* Global Variables */

FILE *DDL_ptr;

extern int No_key_cols, No_auto_relns, No_man_relns;
extern int Def_key_vals;

extern char Table_name[33], Dbproc_name[DB_OBJ_LEN_S];
extern char Rdate[25], Username[25];

extern struct obj_name *Relns_auto, *Relns_man, *Man_cols;
extern struct key_info *Keyptr;



/* FUNCTION CODE */


/*
** Generate a reference insert & update database procedure, if necessary.
** The main logic of this INSERT & UPDATE function is that when a row is
** inserted or a key column is updated in a table that has Foreign keys
** to a related table (ie. it is a child), then the Primary key must be
** valid in the related table (ie. the parent).
*/

int ref_dbproc()
{
    int result;
    char ddl_script[FILE_LEN];

    get_name(Table_name, "\0", "_ref", DB_OBJ_LEN, Dbproc_name);
    if (check_existing(Table_name, Dbproc_name, ddl_script, "dbproc") == IGNORE)
	return TRUE;

    printf("\n\nCreating database procedure for Reference Insert & Update ...");
    result = ref_ins_upd_dbproc(ddl_script);
    final_ref(result);

    return result;
}


/* Finalise */

void final_ref(result)
int result;
{
    if (result == TRUE)
	printf("\nCreated ok\n");
    else
	printf("\nDDL script created, but it contains errors\n");

    system("sleep 2");

    return;
}


/* Control creation of ddl sql script file for a DB Procedure */

int ref_ins_upd_dbproc(ddl_script)
char *ddl_script;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    ref_dbp_header();
    ref_dbproc_args();
    ref_dbproc_sql();

    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void ref_dbp_header()
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Procedure:  %s\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.dbproc\n", Dbproc_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database Procedure to maintain referential "\
		     "integrity for tables that\n");
    fprintf(DDL_ptr, "**   have foreign keys to the %s table.\n", Table_name);
    fprintf(DDL_ptr, "**   This will ensure valid Primary key values in\n");
    fprintf(DDL_ptr, "**   table - %s are used for foreign keys\n", Table_name);
    fprintf(DDL_ptr, "**   in the following tables:-\n");

    list_tables();

    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop procedure before creating  */\n\n");
    fprintf(DDL_ptr, "DROP PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create procedure %s  */\n\n", Dbproc_name);
    fprintf(DDL_ptr, "CREATE PROCEDURE %s\n", Dbproc_name);
    fprintf(DDL_ptr, "(\n");

    return;
}


/* Write in the procedure arguments */

void ref_dbproc_args()
{
    int i;
    char type_length[15], null_default[25];
    char tmp_name[33];
    struct key_info *kptr;

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;

	set_type_length(kptr->c_length, kptr->c_datatype, type_length);

	if (kptr->c_null[0] == 'Y')
	    strcpy(null_default, "with null");
	else
	    {
            strcpy(null_default, "not null");
            if (kptr->c_default[0] == 'Y')
                strcat(null_default, " with default");
            else
                strcat(null_default, " not default");
	    }

	fprintf(DDL_ptr, "    %-32s  %-15s  %s,\n", kptr->c_name, type_length,
						    null_default);
	}

    fprintf(DDL_ptr, "    ref_table\t\t\t      varchar(32)      not null "\
		     "not default\n");

    return;
}


/* Create the main logic */

void ref_dbproc_sql()
{
    int i;
    char c;
    struct key_info *kptr;

    fprintf(DDL_ptr, ") =\n");

    fprintf(DDL_ptr, "DECLARE\n");
    fprintf(DDL_ptr, "\terror_number =\tinteger4\tnot null with default;\n");
    fprintf(DDL_ptr, "\trows =\t\tinteger4\tnot null with default;\n");
    fprintf(DDL_ptr, "\tmsg_text =\tvarchar(240)\tnot null with default;\n\n");

    fprintf(DDL_ptr, "BEGIN\n");

    if (Def_key_vals != 3)
	def_val_code("' + ref_table + '");

    fprintf(DDL_ptr, "\n\t/* Check references */\n\n");
    fprintf(DDL_ptr, "\trows = 0;\n\n");
    fprintf(DDL_ptr, "\tSELECT :rows = 1\n");
    fprintf(DDL_ptr, "\tFROM %s\n", Table_name);

    c = '\0';

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;

	if ((No_key_cols - i) == 1)
	    c = ';';

	if (i == 0)
	    fprintf(DDL_ptr, "\tWHERE %s = :%s%c\n", kptr->c_name,
						     kptr->c_name, c);
	else
	    fprintf(DDL_ptr, "\tAND %s = :%s%c\n", kptr->c_name,
						   kptr->c_name, c);
	}

    fprintf(DDL_ptr, "\n\tIF rows = 0\n");
    fprintf(DDL_ptr, "\tTHEN\n");
    fprintf(DDL_ptr, "\t\t/* Reference Primary key does not exist */\n\n");
    fprintf(DDL_ptr, "\t\terror_number = 100002;\n");
    fprintf(DDL_ptr, "\t\tmsg_text = 'You have referenced an entry in the';\n");
    fprintf(DDL_ptr, "\t\tmsg_text = msg_text + ' table - %s - ';\n",
							Table_name);
    fprintf(DDL_ptr, "\t\tmsg_text = msg_text + 'that does not exist. ';\n");
    fprintf(DDL_ptr, "\t\tmsg_text = msg_text + 'The table being inserted';\n");
    fprintf(DDL_ptr, "\t\tmsg_text = msg_text + ' / updated was ' + "\
		     "ref_table + '.';\n");
    fprintf(DDL_ptr, "\t\tRAISE ERROR error_number msg_text;\n");
    fprintf(DDL_ptr, "\tENDIF;\n");

    fprintf(DDL_ptr, "END;\n\\p\\g\n");

    return;
}


/*
** Generate a reference insert and update database rule, if necessary.
** See reference insert and update database procedure comments for details.
*/

void ref_rule(rel_table, ref_pt)
char *rel_table;
int ref_pt;
{
    int result;
    char rule_name[DB_OBJ_LEN_S];
    char ddl_script[FILE_LEN];
    char prefix[DB_OBJ_LEN_S];

    /* Derive a name prefix */

    strcpy(prefix, Table_name);
    trunc_string(prefix, 9);
    strcat(prefix, "_");

    get_name(rel_table, prefix, "_ref", DB_OBJ_LEN, rule_name);
    if (check_existing(rel_table, rule_name, ddl_script, "rule") == IGNORE)
	return;

    printf("\n\nCreating database rule for Reference Insert & Update ...");
    result = ref_ins_upd_rule(rel_table, rule_name, ddl_script, ref_pt);
    final_ref(result);

    return;
}


/* Control creation of ddl sql script file for a DB Rule */

int ref_ins_upd_rule(rel_table, rule_name, ddl_script, ref_pt)
char *rel_table;
char *rule_name;
char *ddl_script;
int ref_pt;
{
    int result;

    /* Open script file */

    ddl_open(ddl_script, "w");

    /* Construct the sql script */

    ref_rule_header(rel_table, rule_name);
    ref_rule_action(rel_table);
    ref_rule_args(rel_table, ref_pt);

    fprintf(DDL_ptr, "\\p\\g\n");
    fclose(DDL_ptr);

    /* Execute the ddl script */

    result = run_ddl(ddl_script);

    return result;
}


/* Prepare header information */

void ref_rule_header(rel_table, rule_name)
char *rel_table;
char *rule_name;
{
    /* Header documentation */

    fprintf(DDL_ptr, "/*\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Rule:  %s\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   File:  %s.rule\n", rule_name);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Author:  %s (via km_express)\n", Username);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Date:  %s\n", Rdate);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "**   Database rule to enable referential "\
		     "integrity between tables:-\n");
    fprintf(DDL_ptr, "**       %s and %s\n", Table_name, rel_table);
    fprintf(DDL_ptr, "**\n");
    fprintf(DDL_ptr, "*******************************************************"\
                     "********************\n");
    fprintf(DDL_ptr, "*/\n\n");

    /* Drop command */

    fprintf(DDL_ptr, "/*  Drop rule before creating  */\n\n");
    fprintf(DDL_ptr, "DROP RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\\p\\g\n\n");

    /* Create command */

    fprintf(DDL_ptr, "/*  Create Rule %s  */\n\n", rule_name);
    fprintf(DDL_ptr, "CREATE RULE %s\n", rule_name);
    fprintf(DDL_ptr, "\n");

    return;
}


/* Set up the rule action clause */

void ref_rule_action(rel_table)
char *rel_table;
{
    if (No_key_cols == 1)
	fprintf(DDL_ptr, "\tAFTER INSERT, UPDATE (%s) ON %s\n", Keyptr->c_name,
								rel_table);
    else
	fprintf(DDL_ptr, "\tAFTER INSERT, UPDATE ON %s\n", rel_table);

    fprintf(DDL_ptr, "\n\tEXECUTE PROCEDURE %s(\n", Dbproc_name);

    return;
}


/* Format the arguments to the database procedure */

void ref_rule_args(rel_table, ref_pt)
char *rel_table;
int ref_pt;
{
    int i;
    char *c_name;
    struct key_info *kptr;
    struct obj_name *tmp_ptr;

    for(i = 0; i < No_key_cols; i++)
	{
	kptr = Keyptr + i;

	if (ref_pt == -1)
	    c_name = kptr->c_name;
	else
	    {
	    tmp_ptr = Man_cols + (ref_pt + i);
	    c_name = tmp_ptr->name;
	    }

	fprintf(DDL_ptr, "\t\t\t%s = new.%s,\n", kptr->c_name, c_name);
	}

    fprintf(DDL_ptr, "\t\t\tref_table = '%s')\n", rel_table);

    return;
}
