/*
** This code is the property of Waratah Computing Services - 22/10/93.
** It contains Proprietary Information and should be treated as Confidential.
*/

/*
** Functions: 	SQL access routines
**
** Description: Routines to check if something exists or to look up some
**		specific information. These would be better written in
**		ESQL/C, but not everyone has it (stingy old Ingres).
**
** Returns:	TRUE - ok / success
**		FALSE - not exists / DB error
*/

/* Headers */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


/* Macro Definitions */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef DELIM
#define DELIM '|'
#endif

#ifndef BUFFER_LEN
#define BUFFER_LEN 100
#endif

#ifndef FILE_LEN
#define FILE_LEN 256
#endif

#define FIELD_LEN_S 33


/* Function Prototypes */

int sql_lookup(char *, char *);
void sql_script(char *, char *);
int check_found();
int sql_related();
void lookup_keys();
void lookup_cols(char *);
int save_keys();
int save_cols();
void setup_key(char *, int);
void setup_col(char *, int);
int number_rows(char *);
void get_field(char **, char *);
int error_check();
void sql_error(char *);
int run_ddl(char *);

extern void main_final();
extern void append_file(char *, char *);


/* Structures */

extern struct key_info
{
char c_name[33];
char c_datatype[33];
int c_length;
char c_null[9];
char c_default[9];
};

extern struct col_info
{
char c_name[33];
char c_datatype[33];
};


/* Global Variables */

extern FILE *I_PTR;
extern FILE *O_PTR;

extern int No_key_cols, No_columns, Verbose, Def_key_vals;

extern char Restriction[500], Dbname[33], Table_name[33], Username[25];
extern char Log_file[FILE_LEN];
extern char *Sql_in, *Sql_out;

extern struct key_info *Keyptr;
extern struct col_info *Colptr;


/* FUNCTION CODE */


/* General lookup routine to test if a database object exists */

int sql_lookup(db_name, table_name)
char *db_name;
char *table_name;
{
    int result;

    if (Verbose)
	printf("\nProgram sql_lookup starting...");

    sql_script(db_name, table_name);
    result = check_found();

    if (Verbose)
	printf("\nProgram sql_lookup completed\n");

    return result;
}


/* Prepare the sql script and execute */

void sql_script(db_name, table_name)
char *db_name;
char *table_name;
{
    int rc;
    char sys_cmd[100];

    /* Open the file */

    if ((I_PTR = fopen(Sql_in, "w")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql script file - %s", Sql_in);
	exit(1);
	}

    /* Set up sql commands */

    fprintf(I_PTR, "select 'OBJ_FOUN' + 'D' from %s\n", table_name);
    fprintf(I_PTR, "where %s;\n", Restriction);
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\\q\n");

    fclose(I_PTR);

    /* Run the sql */

    sprintf(sys_cmd, "sql %s < %s > %s 2> /dev/null", db_name, Sql_in, Sql_out);

    if ((rc = system(sys_cmd)) != 0)
	{
	printf("\n\nERROR: Unable to run command - %s\n", sys_cmd);
	main_final();
	exit(1);
	}

    return;
}


int check_found()
{
    int found_ind;
    char buffer[BUFFER_LEN];
    char *p;

    /* Open the file */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
	exit(1);
	}

    /* Read the file and check for errors or database found */

    found_ind = FALSE;
    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "OBJ_FOUND")) != (char *) NULL)
	    {
	    found_ind = TRUE;
	    break;
	    }

	if ((p = strstr(buffer, "E_")) != (char *) NULL)
	    sql_error(buffer);
	}

    fclose(O_PTR);

    return found_ind;
}


/* Find (possible) related tables for a table */

int sql_related()
{
    int rc;
    char sys_cmd[100];

    /* Open the file */

    if ((I_PTR = fopen(Sql_in, "w")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql script file - %s", Sql_in);
	exit(1);
	}

    /* Set up sql commands */

    fprintf(I_PTR, "select\tc2.table_name,\n");
    fprintf(I_PTR, "\tcount(c2.table_name),\n");
    fprintf(I_PTR, "\t'OBJ_FOUN' + 'D'\n");
    fprintf(I_PTR, "from iicolumns c1, iicolumns c2, iitables t\n");
    fprintf(I_PTR, "where c1.key_sequence > 0\n");
    fprintf(I_PTR, "and c1.table_name = '%s'\n", Table_name);
    fprintf(I_PTR, "and c2.table_name != '%s'\n", Table_name);
    fprintf(I_PTR, "and c1.table_owner = '%s'\n", Username);
    fprintf(I_PTR, "and c2.table_owner = '%s'\n", Username);
    fprintf(I_PTR, "and c2.table_name = t.table_name\n");
    fprintf(I_PTR, "and t.table_owner = '%s'\n", Username);
    fprintf(I_PTR, "and t.table_type = 'T'\n");
    fprintf(I_PTR, "and c1.column_name = c2.column_name\n");
    fprintf(I_PTR, "group by c2.table_name\n");
    fprintf(I_PTR, "having count(c2.table_name) = %d\n", No_key_cols);
    fprintf(I_PTR, "order by 1;\n");
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\\q\n");

    fclose(I_PTR);

    /* Run the sql */

    sprintf(sys_cmd, "sql %s < %s > %s 2> /dev/null", Dbname, Sql_in, Sql_out);

    if ((rc = system(sys_cmd)) != 0)
	{
	printf("\n\nERROR: Unable to run command - %s\n", sys_cmd);
	return FALSE;
	}

    if ((rc = error_check()) == FALSE)
	return FALSE;

    return TRUE;
}


/* Look up the Primary key column(s) for a table */

void lookup_keys()
{
    int rc;
    char sys_cmd[100];

    /* Open the sql input file */

    if ((I_PTR = fopen(Sql_in, "w")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql script file - %s", Sql_in);
	exit(1);
	}

    /* Set up sql commands */

    fprintf(I_PTR, "select max(key_sequence), 'NUM_COL' + 'S'\n");
    fprintf(I_PTR, "from iicolumns\n");
    fprintf(I_PTR, "where table_name = '%s'\n", Table_name);
    fprintf(I_PTR, "and table_owner = '%s'\n", Username);
    fprintf(I_PTR, "and key_sequence > 0;\n");
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\n\n");

    fprintf(I_PTR, "select 'COL_FOUN' + 'D', column_name,\n");
    fprintf(I_PTR, "       lowercase(column_datatype),\n");
    fprintf(I_PTR, "       column_length, column_nulls, column_defaults,\n");
    fprintf(I_PTR, "       key_sequence\n");
    fprintf(I_PTR, "from iicolumns\n");
    fprintf(I_PTR, "where table_name = '%s'\n", Table_name);
    fprintf(I_PTR, "and table_owner = '%s'\n", Username);
    fprintf(I_PTR, "and key_sequence > 0\n");
    fprintf(I_PTR, "order by key_sequence;\n");
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\\q\n");

    fclose(I_PTR);

    /* Run the sql */

    sprintf(sys_cmd, "sql %s < %s > %s 2> /dev/null", Dbname, Sql_in, Sql_out);

    if ((rc = system(sys_cmd)) != 0)
	{
	printf("\n\nERROR: Unable to run command - %s\n", sys_cmd);
	main_final();
	exit(1);
	}

    return;
}


/* Look up the Mandatory column(s) for a table */

void lookup_cols(table)
char *table;
{
    int rc;
    char sys_cmd[100];

    /* Open the sql input file */

    if ((I_PTR = fopen(Sql_in, "w")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql script file - %s", Sql_in);
	exit(1);
	}

    /*
    ** Set up sql commands
    ** Do not create for key columns if the DEF_KVALS env. variable
    ** is set to 2 as the check will already be taken care in other rules.
    */

    fprintf(I_PTR, "select count(column_name), 'NUM_COL' + 'S'\n");
    fprintf(I_PTR, "from iicolumns\n");
    fprintf(I_PTR, "where table_name = '%s'\n", table);
    fprintf(I_PTR, "and table_owner = '%s'\n", Username);

    if (Def_key_vals == 2)
	fprintf(I_PTR, "and key_sequence = 0\n");

    fprintf(I_PTR, "and column_defaults = 'N'\n");
    fprintf(I_PTR, "and column_nulls = 'N';\n");
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\n\n");

    fprintf(I_PTR, "select 'COL_FOUN' + 'D', column_name,\n");
    fprintf(I_PTR, "       lowercase(column_datatype)\n");
    fprintf(I_PTR, "from iicolumns\n");
    fprintf(I_PTR, "where table_name = '%s'\n", table);
    fprintf(I_PTR, "and table_owner = '%s'\n", Username);

    if (Def_key_vals == 2)
	fprintf(I_PTR, "and key_sequence = 0\n");

    fprintf(I_PTR, "and column_defaults = 'N'\n");
    fprintf(I_PTR, "and column_nulls = 'N'\n");
    fprintf(I_PTR, "order by column_name;\n");
    fprintf(I_PTR, "commit;\n");
    fprintf(I_PTR, "\\p\\g\\q\n");

    fclose(I_PTR);

    /* Run the sql */

    sprintf(sys_cmd, "sql %s < %s > %s 2> /dev/null", Dbname, Sql_in, Sql_out);

    if ((rc = system(sys_cmd)) != 0)
	{
	printf("\n\nERROR: Unable to run command - %s\n", sys_cmd);
	main_final();
	exit(1);
	}

    return;
}


/* Save the Primary key column(s) for a table */

int save_keys()
{
    int i;
    char buffer[BUFFER_LEN];
    char *p;

    /* Open the sql output file */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
	exit(1);
	}

    /* Read the file and check for errors or database found */

    i = 0;

    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "COL_FOUND")) != (char *) NULL)
	    {
	    setup_key(buffer, i);
	    i++;
	    continue;
	    }

	if ((p = strstr(buffer, "NUM_COLS")) != (char *) NULL)
	    {
	    if ((No_key_cols = number_rows(buffer)) == 0)
		return FALSE;
	    else
		{
		Keyptr = (struct key_info *)
		 	 calloc(No_key_cols, sizeof(struct key_info));

		if (Keyptr == NULL)
		    {
		    printf("\n\nERROR: No space available for key storage\n");
		    main_final();
		    exit(1);
		    }
		}
	    }

	if ((p = strstr(buffer, "E_")) != (char *) NULL)
	    sql_error(buffer);
	}

    fclose(O_PTR);

    return TRUE;
}


/* Save the details of a key column */

void setup_key(s, i)
char *s;
int i;
{
    char field[FIELD_LEN_S];
    char *p;
    struct key_info *tmp_ptr;

    /* Skip the first field */

    p = s;
    p++;

    if ((p = strchr(p, DELIM)) == (char *) NULL)
	{
	printf("\n\nERROR: SQL output is not in a valid format.");
	printf("\n\t       Examine the file - %s\n", Sql_out);
	exit(1);
	}

    /* get fields one at a time */

    tmp_ptr = Keyptr + i;

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_name, field);

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_datatype, field);

    p++;
    get_field(&p, field);
    tmp_ptr->c_length = atoi(field);

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_null, field);

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_default, field);

    return;
}


/* Save the Not Null / Not Default column(s) for a table */

int save_cols()
{
    int i;
    char buffer[BUFFER_LEN];
    char *p;

    /* Open the sql output file */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
	exit(1);
	}

    /* Read the file and check for errors or database found */

    i = 0;

    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "COL_FOUND")) != (char *) NULL)
	    {
	    setup_col(buffer, i);
	    i++;
	    continue;
	    }

	if ((p = strstr(buffer, "NUM_COLS")) != (char *) NULL)
	    {
	    if ((No_columns = number_rows(buffer)) == 0)
		return FALSE;
	    else
		{
		Colptr = (struct col_info *)
		 	 calloc(No_columns, sizeof(struct col_info));
		}
	    }

	if ((p = strstr(buffer, "E_")) != (char *) NULL)
	    sql_error(buffer);
	}

    fclose(O_PTR);

    return TRUE;
}


/* Save the details of a mandatory column */

void setup_col(s, i)
char *s;
int i;
{
    char field[FIELD_LEN_S];
    char *p;
    struct col_info *tmp_ptr;

    /* Skip the first field */

    p = s;
    p++;

    if ((p = strchr(p, DELIM)) == (char *) NULL)
	{
	printf("\n\nERROR: SQL output is not in a valid format.");
	printf("\n\t       Examine the file - %s\n", Sql_out);
	exit(1);
	}

    /* get fields one at a time */

    tmp_ptr = Colptr + i;

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_name, field);

    p++;
    get_field(&p, field);
    strcpy(tmp_ptr->c_datatype, field);

    return;
}


/* Return the nth field on a line */

void get_field(p, field)
char **p;
char *field;
{
    int i;

    for(i = 0; **p != DELIM && **p != '\0'; (*p)++)
	{
	if (**p == ' ')
	    continue;

	field[i++] = **p;
	}

    field[i] = '\0';

    return;
}


/* Determine the number of key columns and allocate space */

int number_rows(s)
char *s;
{
    int i, j, nrows;
    char row_char[5];

    /* Isolate row count */

    for(i = 1, j = 0; s[i] != DELIM && s[i] != '\0'; i++)
	{
	if (isdigit(s[i]))
	    row_char[j++] = s[i];
	}

    row_char[j] = '\0';
    nrows = atoi(row_char);

    return nrows;
}


/* Load an sql script */

int run_ddl(ddl_script)
char *ddl_script;
{
    int rc;
    int result = TRUE;
    char sys_cmd[100];
    char buffer[BUFFER_LEN];
    char *p;

    /* Run the sql */

    sprintf(sys_cmd, "sql %s < %s > %s 2>&1", Dbname, ddl_script, Sql_out);

    if ((rc = system(sys_cmd)) != 0)
	{
	printf("\n\nERROR: Unable to run command - %s\n", sys_cmd);
	main_final();
	exit(1);
	}

    /* Open the log file to look for errors */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
	exit(1);
	}

    /* Check for errors in the ddl sql script */

    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "E_US0965")) != (char *) NULL)
	    if ((p = strstr(p, "not found")) != (char *) NULL)
		continue;

	if ((p = strstr(buffer, "E_US189D")) != (char *) NULL)
	    if ((p = strstr(p, "not exist")) != (char *) NULL)
		continue;

	if (buffer[0] == 'E' && buffer[1] == '_')
	    {
	    printf("\n\nERROR:  The DDL script %s contains errors", ddl_script);
	    printf("\n\tFor details examine the file - %s\n", Log_file);
	    result = FALSE;
	    break;
	    }
	}

    fclose(O_PTR);
    append_file(Log_file, Sql_out);

    return result;
}


/* Check for SQL error */

int error_check()
{
    char buffer[BUFFER_LEN];
    char *p;

    /* Open the file */

    if ((O_PTR = fopen(Sql_out, "r")) == NULL)
	{
	printf("\n\nERROR: Unable to open sql output file - %s", Sql_out);
	exit(1);
	}

    /* Read the file and check for errors */

    while ((fgets(buffer, BUFFER_LEN, O_PTR)) != (char *) NULL)
	{
	if ((p = strstr(buffer, "E_")) != (char *) NULL)
	    {
	    sql_error(buffer);
	    break;
	    }
	}

    fclose(O_PTR);

    if (p != (char *) NULL)
	return FALSE;
    else
	return TRUE;
}


/* Error handler */

void sql_error(buffer)
char *buffer;
{
    printf("\n\nERROR: SQL SELECT failed, message details follow.");
    printf("\n\t%s", buffer);
    printf("\n\tFor more information examine the file - %s\n", Sql_out);
    main_final();
    exit(1);
}
